;****************************************************************
;*  DOWN2000.ASM						*
;*			Countdown bis 3. Jahrtausend		*
;*			============================		*
;*								*
;*  04.10.98	Bodo Wenzel	V1.0, Erstellung		*
;****************************************************************

	title		"Countdown 2000"

VERSION	equ	H'10'

;================================================================
;	Definitionen

	processor	16c84
	radix		dec
	include		"p16cxx.inc"

;================================================================
;	Konfiguration

	org	H'2007'
	dw	_CP_OFF & _PWRTE_ON & _WDT_OFF & _RC_OSC

;================================================================
;	Ein- und Ausgnge

TRIS_A	equ	B'00010000'	;Port A:
;	equ	0..3		;O Digit
HZ50	equ	4		;I 50-Hz-Netzpulse

TRIS_B	equ	B'00000000'	;Port B:
;	equ	0..7		;O Segmente

OPT	equ	B'11011111'	;TMR0 als Timer F_CYCLE/1

;================================================================
;	Daten
	org	H'0C'

int_w:		res	1
int_status:	res	1	;Puffer whrend Interrupt

flags:		res	1	;diverse Flags (siehe fl_...)
fl_hz50		equ	0	;erkannter Pegel von HZ50
fl_key		equ	1	;Taste an HZ50 erkannt
fl_enter	equ	2	;Eingabemodus

debounce:	res	2	;Entprellzhler

counter:	res	11	;Zhlerdigits

segments:	res	1
digit:		res	1	;fr Anzeige

calc_count:	res	1
calc_bin:	res	2
calc_dec:	res	2	;Speicher fr Berechnungen

	if	$ > H'30'
	error	"Zu viel RAM benutzt"
	endif

;================================================================
;	Vektoren
	org	H'0000'
	goto	reset

	dw	VERSION

	org	H'0004'		;geht ohne Sprung weiter...
;================================================================
;	Interrupts
interrupt:
	movwf	int_w
	swapf	STATUS,w
	bcf	STATUS,RP0
	movwf	int_status	;alles retten

	bcf	INTCON,T0IF	;Timer zurcksetzen
;----------------------------------------------------------------
	incf	digit
	movf	digit,w
	xorlw	10
	btfsc	STATUS,Z
	clrf	digit		;Digit zyklisch weiterzhlen

	movf	digit,w
	addlw	counter
	movwf	FSR
	comf	INDF,w
	call	bin2ssg
	movwf	segments	;Segmente holen

	btfss	flags,fl_enter
	bsf	segments,2
	btfss	digit,3
	btfss	digit,0
	bcf	segments,2	;evtl. Dezimalpunkt setzen

	movlw	15
	movwf	PORTA
	movf	segments,w
	movwf	PORTB
	movf	digit,w
	movwf	PORTA		;nchstes Digit schalten
;----------------------------------------------------------------
	btfsc	PORTA,HZ50
	goto	i_hz50_h	;Pegel "high"?

	incf	debounce+0
	btfss	debounce+0,2
	goto	i_quit		;"low" noch nicht entprellt?

	decf	debounce+0
	bsf	flags,fl_hz50	;"low" erkannt

	incfsz	debounce+1
	goto	i_quit		;Taste noch nicht erkannt?

	decf	debounce+1
	bsf	flags,fl_key	;Taste erkannt
	goto	i_quit

i_hz50_h:
	decf	debounce+0
	btfss	debounce+0,7
	goto	i_quit		;"high" noch nicht erkannt?

	incf	debounce+0
	bcf	flags,fl_hz50	;"high" erkannt
	clrf	debounce+1
	bcf	flags,fl_key	;Taste wieder losgelassen
;----------------------------------------------------------------
i_quit:
	swapf	int_status,w
	movwf	STATUS
	swapf	int_w
	swapf	int_w,w		;alles wiederholen
	retfie
;================================================================
bin2ssg:
	addwf	PCL		;je nach Segment springen
		; fgbac.ed
	retlw	B'10111011'	;0
	retlw	B'00101000'	;1
	retlw	B'01110011'	;2
	retlw	B'01111001'	;3
	retlw	B'11101000'	;4
	retlw	B'11011001'	;5
	retlw	B'11011011'	;6
	retlw	B'00111000'	;7
	retlw	B'11111011'	;8
	retlw	B'11111001'	;9
;================================================================
bin2month:
	addwf	PCL		;je nach Monat springen
	retlw	31		;12/00
	retlw	30		;11
	retlw	31		;10
	retlw	30		;9
	retlw	31		;8
	retlw	31		;7
	retlw	30		;6
	retlw	31		;5
	retlw	30		;4
	retlw	31		;3
	retlw	29		;2
	retlw	31		;1
	retlw	31		;12/99
	retlw	30		;11
	retlw	31		;10
	retlw	30		;9
	retlw	31		;8
	retlw	31		;7
	retlw	30		;6
	retlw	31		;5
	retlw	30		;4
	retlw	31		;3
	retlw	28		;2
	retlw	31		;1
	retlw	31		;12/98
	retlw	30		;11
	retlw	31		;10
	retlw	30		;9
	retlw	31		;8
	retlw	31		;7
	retlw	30		;6
	retlw	31		;5
	retlw	30		;4
	retlw	31		;3
	retlw	28		;2
	retlw	31		;1

	if	($ >> 8) != (bin2ssg >> 8)
	error	"Tabellensprung ber Seitengrenze"
	endif
;================================================================
;	Initialisierung
reset:
	bsf	STATUS,RP0
	movlw	TRIS_A
	movwf	TRISA		;Port A programmieren
	movlw	TRIS_B
	movwf	TRISB		;Port B programmieren
	movlw	OPT
	movwf	OPTION_REG	;Optionen programmieren
	bcf	STATUS,RP0
;----------------------------------------------------------------
	movlw	bin2ssg >> 8
	movwf	PCLATH		;Sprnge vorbereiten

	clrf	flags
	clrf	debounce+0
	clrf	debounce+1
	movlw	-5
	movwf	counter+1
	movlw	-1
	movwf	counter+0
	movwf	counter+2
	movwf	counter+3
	movwf	counter+4
	movwf	counter+5
	movwf	counter+6
	movwf	counter+7
	movwf	counter+8
	movwf	counter+9
	movwf	counter+10
	clrf	digit		;Variablen vorbereiten

	bsf	INTCON,T0IE
	bsf	INTCON,GIE	;Interrupts freigeben
;================================================================
;	Hauptschleife

tick	macro	label
	local	tick_loop,tick_hl,tick_lh,tick_key
tick_hl:
	btfsc	flags,fl_key
	goto	label		;Taste gedrckt?
	btfsc	flags,fl_hz50
	goto	tick_hl		;noch Puls?
tick_lh:
	btfss	flags,fl_hz50
	goto	tick_lh		;noch kein Puls?
	endm

decr	macro	offs,reload,label
	incf	counter+offs,w
	btfsc	STATUS,Z
	movlw	reload
	movwf	counter+offs
	btfss	STATUS,Z
	goto	label		;Digit mit Begrenzung erniedrigen
	endm

nokey	macro
	local	loop
loop:
	btfsc	flags,fl_key
	goto	loop		;Taste erst loslassen
	endm
;----------------------------------------------------------------
main_loop:
	tick	enter_year	;auf Puls oder Taste warten

	decr	10,-5,main_loop	;1/10-Sekunden-Takt

	decr	0,-10,main_loop
	decr	1,-10,main_loop
	decr	2,-6,main_loop
	decr	3,-10,main_loop
	decr	4,-6,main_loop
	decr	5,-10,main_loop
	decr	6,-3,main_loop
	movlw	-4
	movwf	counter+5
	decr	7,-10,main_loop
	decr	8,-10,main_loop
	decr	9,-10,main_loop	;Zeit herunterzhlen

	goto	main_loop
;----------------------------------------------------------------
enter_year:
	bsf	flags,fl_enter	;Eingabemodus

	movlw	-1
	movwf	counter+9
	movlw	-1
	movwf	counter+8
	movlw	-2
	movwf	counter+7
	movlw	-3
	movwf	counter+6
	movlw	-4
	movwf	counter+5
	movlw	-2
	movwf	counter+4
	movlw	-3
	movwf	counter+3
	movlw	-4
	movwf	counter+2
	movlw	-6
	movwf	counter+1
	movlw	-10
	movwf	counter+0	;00-12-31-23-59 setzen

	nokey
ey_loop:
	tick	enter_month	;auf Puls oder Taste warten

	decr	10,-50,ey_loop	;1/1-Sekunden-Takt

	incfsz	counter+8,w
	goto	ey_check
	movlw	-10
	movwf	counter+9
	movlw	-10
	movwf	counter+8
	goto	ey_loop
ey_check:
	movwf	counter+8
	xorlw	-8
	btfss	STATUS,Z
	goto	ey_loop
	movlw	-1
	movwf	counter+9
	movlw	-1
	movwf	counter+8
	goto	ey_loop		;Jahr zyklisch 00-98 zhlen
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_month:
	nokey
emo_loop:
	tick	enter_day	;auf Puls oder Taste warten

	decr	10,-50,emo_loop	;1/1-Sekunden-Takt

	incfsz	counter+7,w
	goto	emo_check
	incf	counter+6
	incfsz	counter+6,w
	goto	emo_loop
	movlw	-2
	movwf	counter+7
	movlw	-3
	movwf	counter+6
	goto	emo_loop
emo_check:
	decr	6,-10,emo_loop
	incf	counter+7
	goto	emo_loop	;Monat zyklisch 12-01 zhlen
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_day:
	nokey
ed_loop:
	tick	enter_hour	;auf Puls oder Taste warten

	decr	10,-50,ed_loop	;1/1-Sekunden-Takt

	incfsz	counter+5,w
	goto	ed_check
	incf	counter+4
	incfsz	counter+4,w
	goto	ed_loop
	movlw	-4
	movwf	counter+5
	movlw	-2
	movwf	counter+4
	goto	ed_loop
ed_check:
	decr	4,-10,ed_loop
	incf	counter+5
	goto	ed_loop		;Tag zyklisch 31-01 zhlen
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_hour:
	nokey
eh_loop:
	tick	enter_minute	;auf Puls oder Taste warten

	decr	10,-50,eh_loop	;1/1-Sekunden-Takt

	decr	2,-10,eh_loop
	decr	3,-3,eh_loop
	movlw	-4
	movwf	counter+2
	goto	eh_loop		;Stunde zyklisch 23-00 zhlen
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_minute:
	nokey
emi_loop:
	tick	enter_ready	;auf Puls oder Taste warten

	decr	10,-50,emi_loop	;1/1-Sekunden-Takt

	decr	0,-10,emi_loop
	decr	1,-6,emi_loop
	goto	emi_loop	;Minute zyklisch 59-00 zhlen
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
enter_ready:
	incf	counter+9,w
	movlw	0
	btfss	STATUS,Z
	movlw	24
	btfss	counter+8,0
	movlw	12		;Offset fr Jahr

	btfsc	counter+7,0
	addlw	10
	addwf	counter+6,w
	addlw	3+1		;Offset fr Monat dazu

	movwf	calc_count
	clrf	calc_bin+0
	clrf	calc_bin+1
er_day_loop:
	movf	calc_count,w
	btfsc	STATUS,Z
	goto	er_goon		;alle Monate gezhlt?

	decf	calc_count
	movf	calc_count,w
	bcf	INTCON,T0IE
	call	bin2month
	bsf	INTCON,T0IE
	addwf	calc_bin+1
	btfsc	STATUS,C
	incf	calc_bin+0	;Summe der Tage bilden
	goto	er_day_loop

er_goon:
	comf	counter+4,w
	btfss	counter+5,0
	addlw	10
	btfss	counter+5,1
	addlw	20
	subwf	calc_bin+1
	btfss	STATUS,C
	decf	calc_bin+0	;Tag im Monat abziehen

	call	bindec
	movf	calc_dec+0,w
	andlw	H'0F'
	xorlw	H'FF'
	movwf	counter+9
	swapf	calc_dec+1,w
	andlw	H'0F'
	xorlw	H'FF'
	movwf	counter+8
	movf	calc_dec+1,w
	andlw	H'0F'
	xorlw	H'FF'
	movwf	counter+7	;Rest Tage in Anzeige

	clrf	calc_bin+0
	comf	counter+2,w
	btfss	counter+3,0
	addlw	10
	btfss	counter+3,1
	addlw	20
	sublw	23
	movwf	calc_bin+1	;Rest Stunden ermitteln

	call	bindec
	swapf	calc_dec+1,w
	andlw	H'0F'
	xorlw	H'FF'
	movwf	counter+6
	movf	calc_dec+1,w
	andlw	H'0F'
	xorlw	H'FF'
	movwf	counter+5	;Rest Stunden in Anzeige

	clrf	calc_bin+0
	comf	counter+0,w
	btfss	counter+1,0
	addlw	10
	btfss	counter+1,1
	addlw	20
	btfss	counter+1,2
	addlw	40
	sublw	59
	movwf	calc_bin+1	;Rest Minuten ermitteln

	call	bindec
	swapf	calc_dec+1,w
	andlw	H'0F'
	xorlw	H'FF'
	movwf	counter+4
	movf	calc_dec+1,w
	andlw	H'0F'
	xorlw	H'FF'
	movwf	counter+3	;Rest Minuten in Anzeige

	movlw	-6
	movwf	counter+2
	movlw	-10
	movwf	counter+1
	movlw	-10
	movwf	counter+0	;Minute vollmachen
er_quit:
	nokey

	bcf	flags,fl_enter	;kein Eingabemodus mehr
	goto	main_loop
;================================================================
;	Binrzahl in BCD umrechnen (nur 3 Stellen Ergebnis!)
bindec:
	movlw	16
	movwf	calc_count
	clrf	calc_dec+0
	clrf	calc_dec+1	;Umrechnung vorbereiten
bd_loop:
	movlw	H'03'
	addwf	calc_dec+0
	btfss	calc_dec+0,3
	subwf	calc_dec+0
	addwf	calc_dec+1
	btfss	calc_dec+1,3
	subwf	calc_dec+1
	movlw	H'30'
	addwf	calc_dec+1
	btfss	calc_dec+1,7
	subwf	calc_dec+1	;BCD-Korrektur

	rlf	calc_bin+1
	rlf	calc_bin+0
	rlf	calc_dec+1
	rlf	calc_dec+0	;ein Bit weiterschieben

	decfsz	calc_count
	goto	bd_loop		;noch nicht alle Bits?
	return
;================================================================
	end
